;; ===========================================================================
;; ODBC FFI Functions for clisp
;; ===========================================================================

(USE-PACKAGE "FFI")

;; ===========================================================================
;; wrap ffi c variables to achieve a common access model
;; ===========================================================================
;; For each Common Lisp implementation the following lines have to be re-
;; written. c structures are not considered yet.
;;
;; usage:
;;
;;  (setf x (make-c-<type>))  ;; creates a c variable of type long
;;                            ;; <type> yet may be int long or string
;;
;;  (%ADR x) ;; is used to give the callee a reference to x if of type long*
;;  (%VAL x) ;; obtains the value of any c variable (yet int long or string)
;;
;; ===========================================================================

(defun make-c-int ()
  (FFI:ALLOCATE-DEEP 'sint16 0))

(defun make-c-long ()
  (FFI:ALLOCATE-DEEP 'long 0))

(defun make-c-string (length)
  (FFI:ALLOCATE-DEEP 'character "" :count length))

(defun make-handle ()
  (make-c-long))

(defmacro %VAL (c-variable)
  `(FFI:FOREIGN-VALUE ,c-variable))

(defmacro %ADR (c-variable)
  `(FFI:FOREIGN-ADDRESS ,c-variable))

;; for compatibilty to gold hill; not nice here :-(

(defmacro string-append (&rest strings)
  `(concatenate 'string ,@strings))

;; ===========================================================================
;; of course the next few lines are implementation dependent
;; ===========================================================================

(FFI:DEF-CALL-OUT SQLAllocHandle
  (:library "odbc32.dll")
  (:language :stdc)
  (:name "SQLAllocHandle")
  (:arguments
    (HandleType sint16)
    (InputHandle long)
    (OutputHandlePtr c-pointer))
  (:return-type sint16))

(FFI:DEF-CALL-OUT SQLBindCol
  (:library "odbc32.dll")
  (:language :stdc)
  (:name "SQLBindCol")
  (:arguments
    (StatementHandle sint32)
    (ColumnNumber sint16)
    (TargetType sint16)
    (TargetValuePtr c-pointer)
    (BufferLength sint32)
    (StrLen_or_IndPtr c-pointer))
  (:return-type sint16))

(FFI:DEF-CALL-OUT SQLConnectI
  (:library "odbc32.dll")
  (:language :stdc)
  (:name "SQLConnect")
  (:arguments
    (ConnectionHandle sint32)
    (ServerName c-string)
    (NameLength1 sint16)
    (UserName c-string)
    (NameLength2 sint16)
    (Authentication c-string)
    (NameLength3 sint16))
  (:return-type sint16))

(defun SQLConnect (ConnectionHandle dsn uid pwd)
  (SQLConnectI ConnectionHandle
               dsn SQL_NTS uid SQL_NTS pwd SQL_NTS))

(FFI:DEF-CALL-OUT SQLDescribeCol
  (:library "odbc32.dll")
  (:language :stdc)
  (:name "SQLDescribeCol")
  (:arguments
    (StatementHandle sint32)
    (ColumnNumber sint16)
    (ColumnNamePtr c-pointer)
    (BufferLength sint32)
    (NameLengthPtr c-pointer)
    (DataTypePtr c-pointer)
    (ColumnSizePtr c-pointer)
    (DecimalDigitsPtr c-pointer)
    (NullablePtr c-pointer))
  (:return-type sint16))

(FFI:DEF-CALL-OUT SQLDisconnect
  (:library "odbc32.dll")
  (:language :stdc)
  (:name "SQLDisconnect")
  (:arguments
    (Handle sint32))
  (:return-type sint16))

(FFI:DEF-CALL-OUT SQLEndTran
  (:library "odbc32.dll")
  (:language :stdc)
  (:name "SQLEndTran")
  (:arguments
    (HandleType sint16)
    (Handle sint32)
    (CompletionType sint16))
  (:return-type sint16))

(FFI:DEF-CALL-OUT SQLExecDirectI
  (:library "odbc32.dll")
  (:language :stdc)
  (:name "SQLExecDirect")
  (:arguments
    (StatementHandle sint32)
    (StatementText c-string)
    (TextLength1 sint16))
  (:return-type sint16))

(defun SQLExecDirect (handle statement)
  (SQLExecDirectI handle statement SQL_NTS))

(FFI:DEF-CALL-OUT SQLFetch
  (:library "odbc32.dll")
  (:language :stdc)
  (:name "SQLFetch")
  (:arguments
    (StatementHandle sint32))
  (:return-type sint16))

(FFI:DEF-CALL-OUT SQLFreeHandle
  (:library "odbc32.dll")
  (:language :stdc)
  (:name "SQLFreeHandle")
  (:arguments
    (HandleType sint16)
    (Handle sint32))
  (:return-type sint16))

(FFI:DEF-CALL-OUT SQLGetDiagRec
  (:library "odbc32.dll")
  (:language :stdc)
  (:name "SQLGetDiagRec")
  (:arguments
    (HandleType sint16)
    (Handle sint32)
    (RecNumber sint16)
    (Sqlstate c-pointer)
    (NativeErrorPtr c-pointer)
    (MessageText c-pointer)
    (BufferLength sint16)
    (TextLengthPtr c-pointer))
  (:return-type sint16))

(FFI:DEF-CALL-OUT SQLNumResultCols
  (:library "odbc32.dll")
  (:language :stdc)
  (:name "SQLNumResultCols")
  (:arguments
    (StatementHandle sint32)
    (ColumnCountPtr c-pointer))
  (:return-type sint16))

(FFI:DEF-CALL-OUT SQLSetEnvAttr
  (:library "odbc32.dll")
  (:language :stdc)
  (:name "SQLSetEnvAttr")
  (:arguments
    (EnvironmentHandle sint32)
    (Attribute sint16)
    (ValuePtr uint32)
    (StringLength sint16))
  (:return-type sint16))

;(defun SQLSetEnvAttr (EnvironmentHandle Attribute ValuePtr)
;  (SQLSetEnvAttrI EnvironmentHandle Attribute ValuePtr SQL_NTS))

;; ===========================================================================



