;; ===========================================================================
;; ODBC FFI Functions for Gold Hill Common Lisp
;; ===========================================================================

;; ===========================================================================
;; wrap ffi c variables to achieve a common access model
;; ===========================================================================
;; For each Common Lisp implementation the following lines have to be re-
;; written. c structures are not considered yet.
;;
;; usage:
;;
;;  (setf x (make-c-<type>))  ;; creates a c variable
;;                            ;; <type> yet may be int long or string
;;
;;  (%ADR x) ;; is used to give the callee a reference to x if of type long*
;;  (%VAL x) ;; obtains the value of any c variable (yet int long or string)
;;
;; ===========================================================================

(ffi:def-foreign-struct (c-int (:constructor mk-c-int))
			(value :type :fixnum))

(ffi:def-foreign-struct (c-long (:constructor mk-c-long))			
			(value :type :signed-32bit))

#||
(ffi:def-foreign-struct (c-string (:constructor mk-c-string))
			(value :type :stringz))
||#

(defstruct c-place type var)

(defun make-c-int ()
  (make-c-place :type 'int    :var (mk-c-int)))

(defun make-c-long ()
  (make-c-place :type 'long   :var (mk-c-long)))

#||
(defun make-c-string (length)
  (declare (ignore length))
  (make-c-place :type 'string :var (mk-c-string)))

(defun make-c-string (length &optional ini)
  (let ((v (make-c-place :type 'string :var (mk-c-string))))
    (unless ini (setf ini (code-char 0)))
    (setf (%val v) (make-string length :initial-element ini))
    v))
||#

(defun make-c-string (length)
  (let ((p (ffi:allocate-foreign-pointer (max 4 length))))
    (make-c-place :type 'string :var p)))

(defun make-handle ()
  (make-c-long))

(defun %VAL (c-var-struct)
  (cond 
        ((equalp (c-place-type c-var-struct) 'int)
         (c-int-value (c-place-var c-var-struct)))
        ((equalp (c-place-type c-var-struct) 'long)
         (c-long-value (c-place-var c-var-struct)))
        ((equalp (c-place-type c-var-struct) 'string)
         (ffi::foreign-string-value (c-place-var c-var-struct)))
        (t nil)
        )
  )

(defun %set-VAL% (c-var-struct val)
  (cond 
        ((equalp (c-place-type c-var-struct) 'int)
         (setf (c-int-value (c-place-var c-var-struct)) val))
        ((equalp (c-place-type c-var-struct) 'long)
         (setf (c-long-value (c-place-var c-var-struct)) val))
        ((equalp (c-place-type c-var-struct) 'string)
         (ffi::write-foreign-data (string-append val (string (code-char 0))) (c-place-var c-var-struct) 0)
         (%VAL c-var-struct))
        (t nil)))

(defsetf %VAL %set-VAL%)

(defun %ADR (c-var-struct)
  (c-place-var c-var-struct))

;; ===========================================================================
;; of course the next few lines are implementation dependent
;; ===========================================================================

(ffi:def-foreign-library "ODBC32" :size 5 :place "odbc32.dll")

(ffi:load-foreign-library "ODBC32")

(ffi:def-foreign-function (SQLAllocHandle (:name "SQLAllocHandle")
                                    (:library "ODBC32")
                                    (:return-type :fixnum))
                           (HandleType :fixnum)
                           (InputHandle :signed-32bit)
                           (OutputHandlePtr :pointer))

(ffi:def-foreign-function (SQLBindCol (:name "SQLBindCol")
                                    (:library "ODBC32")
                                    (:return-type :fixnum))
                           (StatementHandle :signed-32bit)
                           (ColumnNumber :fixnum)
                           (TargetType :fixnum)
                           (TargetValuePtr :pointer)
                           (BufferLength :signed-32bit)
                           (StrLen_or_IndPtr :pointer))

(ffi:def-foreign-function (SQLBindParameter (:name "SQLBindParameter")
                                    (:library "ODBC32")
                                    (:return-type :fixnum))
                           (StatementHandle :signed-32bit)
                           (ParameterNumber :fixnum)
                           (InputOutputType :fixnum)
                           (ValueType :fixnum)
                           (ParameterType :fixnum)
                           (ColumnSize :fixnum)
                           (DecimalDigits :fixnum)
                           (ParameterValuePtr :pointer)
                           (BufferLength :signed-32bit)
                           (StrLen_or_IndPtr :pointer))

(ffi:def-foreign-function (SQLConnectI (:name "SQLConnect")
                                    (:library "ODBC32")
                                    (:return-type :fixnum))
                           (ConnectionHandle :signed-32bit)
                           (ServerName :stringz)
                           (NameLength1 :fixnum)
                           (UserName :stringz)
                           (NameLength2 :fixnum)
                           (Authentication :stringz)
                           (NameLength3 :fixnum))

(defun SQLConnect (handle dsn uid pwd)
  (SQLConnectI handle
               dsn (length dsn)
               uid (length uid)
               pwd (length pwd)))

(ffi:def-foreign-function (SQLDescribeCol (:name "SQLDescribeCol")
                                    (:library "ODBC32")
                                    (:return-type :fixnum))
                           (StatementHandle :signed-32bit)
                           (ColumnNumber :fixnum)
                           (ColumnNamePtr :pointer)
                           (BufferLength :signed-32bit)
                           (NameLengthPtr :pointer)
                           (DataTypePtr :pointer)
                           (ColumnSizePtr :pointer)
                           (DecimalDigitsPtr :pointer)
                           (NullablePtr :pointer))

(ffi:def-foreign-function (SQLDisconnect (:name "SQLDisconnect")
                                    (:library "ODBC32")
                                    (:return-type :fixnum))
                           (Handle :signed-32bit))

(ffi:def-foreign-function (SQLEndTran (:name "SQLEndTran")
                                    (:library "ODBC32")
                                    (:return-type :fixnum))
                           (HandleType :fixnum)
                           (Handle :signed-32bit)
                           (CompletionType :fixnum))

(ffi:def-foreign-function (SQLExecDirectI (:name "SQLExecDirect")
                                    (:library "ODBC32")
                                    (:return-type :fixnum))
                           (StatementHandle :signed-32bit)
                           (StatementText :stringz)
                           (TextLength1 :fixnum))

(defun SQLExecDirect (handle statement)
  (SQLExecDirectI handle statement SQL_NTS))

(ffi:def-foreign-function (SQLFetch (:name "SQLFetch")
                                    (:library "ODBC32")
                                    (:return-type :fixnum))
                           (StatementHandle :signed-32bit))

(ffi:def-foreign-function (SQLFreeHandle (:name "SQLFreeHandle")
                                    (:library "ODBC32")
                                    (:return-type :fixnum))
                           (HandleType :fixnum)
                           (Handle :signed-32bit))

(ffi:def-foreign-function (SQLGetDiagRec (:name "SQLGetDiagRec")
                                    (:library "ODBC32")
                                    (:return-type :fixnum))
                           (HandleType :fixnum)
                           (Handle :signed-32bit)
                           (RecNumber :fixnum)
                           (Sqlstate :pointer)
                           (NativeErrorPtr :pointer)
                           (MessageText :pointer)
                           (BufferLength :fixnum)
                           (TextLengthPtr :pointer))


(ffi:def-foreign-function (SQLNumResultCols (:name "SQLNumResultCols")
                                    (:library "ODBC32")
                                    (:return-type :fixnum))
                           (StatementHandle :signed-32bit)
                           (ColumnCountPtr :pointer))

(ffi:def-foreign-function (SQLSetConnectAttr (:name "SQLSetConnectAttr")
                                    (:library "ODBC32")
                                    (:return-type :fixnum))
                           (ConnectionHandle :signed-32bit)
                           (Attribute :fixnum)
                           (ValuePtr :unsigned-32bit)
                           (StringLength :fixnum))

(ffi:def-foreign-function (SQLSetEnvAttr (:name "SQLSetEnvAttr")
                                    (:library "ODBC32")
                                    (:return-type :fixnum))
                           (EnvironmentHandle :signed-32bit)
                           (Attribute :fixnum)
                           (ValuePtr :unsigned-32bit)
                           (StringLength :fixnum))

;; ===========================================================================


